(* Modular math ops
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(* Modules that allows functors that do generic math without needing a specific type *)

(* This describes the type of math. *)
type integer
type real

module type Ops = sig
  type t
  type num_type
  val add: t -> t -> t
  val sub: t -> t -> t
  val succ: t -> t
  val pred: t -> t
  val mul: t -> t -> t
  val div: t -> t -> t
  val rem: t -> t -> t
  val abs: t -> t
  val zero: t
  val one: t
  val min: t
  val max: t
  val print: out_channel -> t -> unit
  val to_float: t -> float
  val of_float: float -> t
end

module IntOps : (Ops with type num_type = integer and type t = int)
module Int32Ops : (Ops with type num_type = integer and type t = int32)
module Int64Ops : (Ops with type num_type = integer and type t = int64)
module NativeOps : (Ops with type num_type = integer and type t = nativeint)
module FloatOps : (Ops with type num_type = real and type t = float)

(* Useful functions *)

(* [succ_float n] returns $n+1.0$ and [pred_float n] returns $n-1.0$ *)
val succ_float: float -> float
val pred_float: float -> float

(* Like compare, but takes an extra epsilon value to use in figuring
out if the floats are 'close enough' to be considered equal. See The
Art of Computer Programming, 4.2.2. As an example,
[fcmp ~epsilon:0.00001 5.000005 5.000006] returns 0, meaning
$5.000005 \sim 5.000006$. *)

val fcmp: epsilon:float -> float -> float -> int

(* Useful constants *)
val e: float
val pi: float
