(* Mersenne Twister PRNG for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(* Ocaml version of the Mersenne Twister random number generator based
   on mt19937ar.c.

   http://www.math.keio.ac.jp/matumoto/emt.html

*)

(* Used to get more space-efficient storage of the data array *)
open Bigarray

(* Period parameters *)
let n = 624

let m = 397

(* Tempering parameters *)
(* Ocaml really needs a way to write literal int32's and int64's *)
let mt = Array1.create Bigarray.int32 c_layout n
and mask_b = Int32.of_string "0x9d2c5680"
and mask_c = Int32.of_string "0xefc60000" 
and upper_mask = Int32.of_string "0x80000000" (* most significant w-r bits *)
and lower_mask = Int32.of_string "0x7fffffff" (* least significant r bits *)
and mag01 = [| Int32.zero; Int32.of_string "0x9908b0df" |] 
and mask_high = Int32.of_string "0xffff0000"
and seeded = ref false
and mti = ref n

let init32 seed = 
  let m1 = Int32.of_string "1812433253" 
  and mfill = Int32.of_string "0xffffffff" in
    Array1.set mt 0 (Int32.logand seed mfill);
    for i = 1 to n - 1 do
      let mtprev = Array1.get mt (i - 1) in
        Array1.set mt i (Int32.add
                           (Int32.mul m1
                              (Int32.logxor mtprev
                                 (Int32.shift_right_logical mtprev 30)))
                              (Int32.of_int i))
    done;
    seeded := true

let init seed = init32 (Int32.of_int seed)

let init_array32 seed =
  init32 (Int32.of_string "19650218");
  let k = ref (max n (Array.length seed))
  and i = ref 1 
  and j = ref 0 
  and mval1 = Int32.of_string "1664525" 
  and mval2 = Int32.of_string "1566083941" 
  in
    while !k > 0 do
      let mtprev = Array1.get mt (!i - 1) in
        Array1.set mt !i
          (Int32.add
             (Int32.add
                (Int32.logxor (Array1.get mt !i)
                   (Int32.mul
                      (Int32.logxor mtprev
                         (Int32.shift_right_logical mtprev 30))
                      mval1))
                seed.(!j))
             (Int32.of_int !j));
        incr i;
        incr j;
        if !i >= n then begin
          Array1.set mt 0 (Array1.get mt (n - 1));
          i := 1
        end;
        if !j >= Array.length seed then j := 0;
        decr k
    done;

    k := n - 1;

    while !k > 0 do
      let mtprev = Array1.get mt (!i - 1) in
        Array1.set mt !i
          (Int32.sub
           (Int32.logxor
              (Array1.get mt !i)
              (Int32.mul
                 (Int32.logxor
                    mtprev
                    (Int32.shift_right_logical mtprev 30))
                    mval2))
              (Int32.of_int !i));
        incr i;
        if !i >= n then begin
          Array1.set mt 0 (Array1.get mt (n - 1));
          i := 1
        end;
        decr k
    done;
    Array1.set mt 0 (Int32.of_string "0x80000000")

let init_bigarray32 seed =
  init32 (Int32.of_string "19650218");
  let k = ref (max n (Array1.dim seed))
  and i = ref 1 
  and j = ref 0 
  and mval1 = Int32.of_string "1664525" 
  and mval2 = Int32.of_string "1566083941" 
  in
    while !k > 0 do
      let mtprev = Array1.get mt (!i - 1) in
        Array1.set mt !i
          (Int32.add
             (Int32.add
                (Int32.logxor (Array1.get mt !i)
                   (Int32.mul
                      (Int32.logxor mtprev
                         (Int32.shift_right_logical mtprev 30))
                      mval1))
                (Array1.get seed !j))
             (Int32.of_int !j));
        incr i;
        incr j;
        if !i >= n then begin
          Array1.set mt 0 (Array1.get mt (n - 1));
          i := 1
        end;
        if !j >= Array1.dim seed then j := 0;
        decr k
    done;

    k := n - 1;

    while !k > 0 do
      let mtprev = Array1.get mt (!i - 1) in
        Array1.set mt !i
          (Int32.sub
           (Int32.logxor
              (Array1.get mt !i)
              (Int32.mul
                 (Int32.logxor
                    mtprev
                    (Int32.shift_right_logical mtprev 30))
                    mval2))
              (Int32.of_int !i));
        incr i;
        if !i >= n then begin
          Array1.set mt 0 (Array1.get mt (n - 1));
          i := 1
        end;
        decr k
    done;
    Array1.set mt 0 (Int32.of_string "0x80000000")

let init_genarray map seed =
  let seed32 = Array.map map seed in
    init_array32 seed32

let init_array seed = init_genarray Int32.of_int seed

let init_from_descr fd =
  let tmp = Array1.create Bigarray.int32 c_layout n
  and t = Array1.map_file fd Bigarray.int32 c_layout false n in
    Array1.blit t tmp;
    init_bigarray32 tmp

let init_from_channel ic =
  let fd = Unix.descr_of_in_channel ic in
    init_from_descr fd
    
let init_from_file fname =
  let fd = Unix.openfile fname [ Unix.O_RDONLY ] 0 in
  let _ = init_from_descr fd in
    Unix.close fd

(* Array1.map_file doesn't work with /dev/urandom, Bah. This is a workaround. *)
let urandom_init () =
  let i = Unix.openfile "/dev/urandom" [ Unix.O_RDONLY ] 0 
  and buf = String.create (4 * n) 
  and ofname = Filename.temp_file "mt_rand" "tmp" in
  let o = Unix.openfile ofname [ Unix.O_RDWR; Unix.O_CREAT ] 0 in
    Unix.unlink ofname;
    let r = Unix.read i buf 0 (String.length buf) in
      if r != String.length buf then
        raise (Failure
                 "Couldn't read the required amount of bytes from /dev/urandom");
      Unix.close i;
      let w = Unix.write o buf 0 r in
        if w != String.length buf then
          raise (Failure
                   "Couldn't write the required amount of bytes to temp file");
      let _ = Unix.lseek o 0 Unix.SEEK_SET in
        init_from_descr o;
        Unix.close o

let urandom_found () =
  try
    let s = Unix.stat "/dev/urandom" in
      s.Unix.st_kind = Unix.S_CHR
  with _ -> false

let self_init () =
  if urandom_found () then urandom_init ()
  else let t = Time.time ()
       and pid = Int32.of_int (Unix.getpid ()) in
    init32 (Int32.logxor t pid)
      
let fill_mt () =
  if not !seeded then self_init ();
  let y = ref Int32.zero 
  and kk = ref 0 in 
  let fiddle i ip ig =
    y := Int32.logor
      (Int32.logand (Array1.get mt i) upper_mask)
      (Int32.logand (Array1.get mt ip) lower_mask);
    Array1.set mt i
      (Int32.logxor
         (Int32.logxor
            (Array1.get mt ig)
            (Int32.shift_right_logical !y 1))
         (Array.unsafe_get mag01 (Int32.to_int (Int32.logand !y Int32.one))))
  in
    while !kk < n - m do
      fiddle !kk (!kk + 1) (!kk + m);
      incr kk
    done;
    while !kk < n - 1 do
      fiddle !kk (!kk + 1) (!kk + m - n);
      incr kk
    done;
    fiddle (n - 1) 0 (m - 1);
    mti := 0

let uint32 () =
  if !mti >= n then fill_mt ();
  let y = Array1.get mt !mti in
  let y' = Int32.logxor y (Int32.shift_right_logical y 11) in
  let y = Int32.logxor y' (Int32.logand (Int32.shift_left y' 7)
                             mask_b) in
  let y' = Int32.logxor y (Int32.logand (Int32.shift_left y 15)
                             mask_c) in
  let r = Int32.logxor y' (Int32.shift_right_logical y' 18) in
    incr mti;
    r
     
let int32 () =
  let r = uint32 () in
    Int32.shift_right_logical r 1
 
let uint64 () =
  let high = Int64.of_int32 (uint32 ())
  and low = Int64.of_int32 (uint32 ()) in
    Int64.logor low (Int64.shift_left high 32) 

let int64 () =
  let r = uint64 () in
    Int64.shift_right_logical r 1

let unativeint () =
  let v = Nativeint.of_int32 (uint32 ()) in
    if Sys.word_size = 32 then v
  else Nativeint.logor v
    (Nativeint.shift_left (Nativeint.of_int32 (uint32 ())) 32)
  
let nativeint () =
  let r = unativeint () in
    Nativeint.shift_right_logical r 1

let uint () =
  let v = unativeint () in
    Nativeint.to_int (Nativeint.shift_right_logical v 1)

let int () =
  let r = unativeint () in
    Nativeint.to_int (Nativeint.shift_right_logical r 2)

let uint32_to_float ui32 =
  if ui32 >= Int32.zero then
    Int32.to_float ui32
  else (* ACK! *) 
    float_of_string (Printf.sprintf "%lu" ui32)

let real0 () =
  let i32 = uint32 () in
    uint32_to_float i32

let real1 () =
  let r = real0 () in
    r *. (1.0 /. 4294967295.0)

let real2 () =
  let r= real0 () in
    r *. (1.0 /. 4294967296.0)

let real3 () =
  let r = real0 () in
    (r +. 0.5) *. (1.0 /. 4294967296.0)

let res53 () =
  let a = Int32.shift_right_logical (uint32 ()) 5 in
  let b = Int32.shift_right_logical (uint32 ()) 6 in
  let a' = uint32_to_float a
  and b' = uint32_to_float b in
    (a' *. 67108864.0 +. b') *. (1.0 /. 9007199254740992.0)

module IntSource = struct
  type t = int
  let genrand () = int ()
  let min = 0
  let max = max_int
end

module Int32Source = struct
  type t = int32
  let genrand () = int32 ()
  let min = Int32.zero
  let max = Int32.max_int
end

module FloatSource = struct
  type t = float
  let genrand () = real1 ()
  let min = 0.0
  let max = 1.0
end
