(* Mersenne Twister PRNG routines for ocaml
   Copyright (C) 2002 Shawn Wagner <raevnos@pennmush.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.
	 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
	 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)

(* Ocaml version of the Mersenne Twister random number generator based
   on mt19937ar.c. See %
\href{http://www.math.keio.ac.jp/matumoto/emt.html}%
{http://www.math.keio.ac.jp/matumoto/emt.html}

*)

(* The seed functions. Only one call to one of these is needed. This should
 * be enough variety. *)

(* Seed from a single value *)
val init32: int32 -> unit
val init: int -> unit

(* Seed from an array of any length *)
val init_array32: int32 array -> unit
val init_array: int array -> unit
val init_bigarray32: (int32, Bigarray.int32_elt, 'a) Bigarray.Array1.t -> unit
val init_genarray: ('a -> int32) -> 'a array -> unit

(* Seed from raw data in a file.
 * The files used should hold at least 700 ints *)
val init_from_file : string -> unit
val init_from_channel: in_channel -> unit
val init_from_descr: Unix.file_descr -> unit

(* Seed from \texttt{/dev/urandom} *)
val urandom_found: unit -> bool
val urandom_init: unit -> unit

(* Seed from \texttt{/dev/urandom}, or a default seed based on time and pid if
 *  \texttt{/dev/urandom} isn't available 
 *)
val self_init: unit -> unit

(* Functions returning random numbers, of the given types and in the
given ranges. All of these but the int64 ones produce the same output
as the reference C code. *)


(* Range: $0\leq x\leq\texttt{0xffffffff}$ *)
val uint32: unit -> int32

(* Range: $0\leq x\leq\texttt{0x7fffffff}$ *)
val int32: unit -> int32

(* Range: $0\leq x\leq\texttt{0xffffffffffffffff}$ *)
val uint64: unit -> int64

(* Range: $0\leq x\leq\texttt{0x7fffffffffffffff}$ *)
val int64: unit -> int64

(* Whichever of the above two is appropriate *)
val unativeint: unit -> nativeint
val nativeint: unit -> nativeint

(* Range: $0\leq x\leq\texttt{0x7fffffff}$ or $0\leq x \leq\texttt{0x7fffffffffffffff}$ *)
val uint: unit -> int

(* Range: $0\leq x\leq\texttt{0x3fffffff}$ or $0\leq x\leq\texttt{0x3fffffffffffffff}$ *)
val int: unit -> int

(* Range: $0\leq x \leq 1$ *)
val real1: unit -> float

(* Range: $0\leq x < 1$ *)
val real2: unit -> float

(* Range: $0<x<1$ *)
val real3: unit -> float

(* Range: $0\leq x<1$ with 53-bit resolution *)

val res53: unit -> float

(* Sources for use with the \textbf{Rand} distributions *)
module IntSource : (Rand.RNGSource with type t = int)
module Int32Source : (Rand.RNGSource with type t = int32)
module FloatSource : (Rand.RNGSource with type t = float)
